"""Tests for ORF finding functionality."""

import pytest

from genome_entropy.orf.types import OrfRecord


def test_orf_record_creation() -> None:
    """Test creating a valid OrfRecord."""
    orf = OrfRecord(
        parent_id="seq1",
        orf_id="seq1_orf_0_90_+_f0",
        start=0,
        end=90,
        strand="+",
        frame=0,
        nt_sequence="A" * 90,
        aa_sequence="K" * 30,  # Added aa_sequence field
        table_id=11,
        has_start_codon=True,
        has_stop_codon=True,
    )
    
    assert orf.parent_id == "seq1"
    assert orf.orf_id == "seq1_orf_0_90_+_f0"
    assert orf.start == 0
    assert orf.end == 90
    assert orf.strand == "+"
    assert orf.frame == 0
    assert len(orf.nt_sequence) == 90
    assert len(orf.aa_sequence) == 30
    assert orf.table_id == 11
    assert orf.has_start_codon is True
    assert orf.has_stop_codon is True
    assert orf.in_genbank is False  # Default value


def test_orf_record_invalid_strand() -> None:
    """Test that invalid strand raises ValueError."""
    with pytest.raises(ValueError, match="Invalid strand"):
        OrfRecord(
            parent_id="seq1",
            orf_id="orf1",
            start=0,
            end=90,
            strand="*",  # Invalid
            frame=0,
            nt_sequence="A" * 90,
            aa_sequence="K" * 30,
            table_id=11,
            has_start_codon=True,
            has_stop_codon=True,
        )


def test_orf_record_invalid_frame() -> None:
    """Test that invalid frame raises ValueError."""
    with pytest.raises(ValueError, match="Invalid frame"):
        OrfRecord(
            parent_id="seq1",
            orf_id="orf1",
            start=0,
            end=90,
            strand="+",
            frame=4,  # Invalid (must be 0, 1, 2, or 3)
            nt_sequence="A" * 90,
            aa_sequence="K" * 30,
            table_id=11,
            has_start_codon=True,
            has_stop_codon=True,
        )


def test_orf_record_invalid_coordinates() -> None:
    """Test that invalid coordinates raise ValueError."""
    # Start < 0
    with pytest.raises(ValueError, match="Invalid start position"):
        OrfRecord(
            parent_id="seq1",
            orf_id="orf1",
            start=-1,
            end=90,
            strand="+",
            frame=0,
            nt_sequence="A" * 91,
            aa_sequence="K" * 30,
            table_id=11,
            has_start_codon=True,
            has_stop_codon=True,
        )
    
    # End <= start
    with pytest.raises(ValueError, match="Invalid end position"):
        OrfRecord(
            parent_id="seq1",
            orf_id="orf1",
            start=90,
            end=90,
            strand="+",
            frame=0,
            nt_sequence="",
            aa_sequence="",
            table_id=11,
            has_start_codon=True,
            has_stop_codon=True,
        )


def test_orf_record_sequence_length_mismatch() -> None:
    """Test that sequence length can be different from coordinates.
    
    Note: The OrfRecord allows sequences to be empty or shorter than coordinates
    because sequences may be filled in later (e.g., by the ORF finder).
    This test now verifies that mismatched lengths are allowed.
    """
    # This should NOT raise an error
    orf = OrfRecord(
        parent_id="seq1",
        orf_id="orf1",
        start=0,
        end=90,
        strand="+",
        frame=0,
        nt_sequence="A" * 60,  # Different length than coordinates
        aa_sequence="K" * 20,
        table_id=11,
        has_start_codon=True,
        has_stop_codon=True,
    )
    assert len(orf.nt_sequence) == 60
    assert orf.end - orf.start == 90


def test_orf_record_both_strands() -> None:
    """Test creating ORFs on both strands."""
    orf_plus = OrfRecord(
        parent_id="seq1",
        orf_id="orf_plus",
        start=0,
        end=90,
        strand="+",
        frame=0,
        nt_sequence="A" * 90,
        aa_sequence="K" * 30,
        table_id=11,
        has_start_codon=True,
        has_stop_codon=True,
    )
    
    orf_minus = OrfRecord(
        parent_id="seq1",
        orf_id="orf_minus",
        start=100,
        end=190,
        strand="-",
        frame=1,
        nt_sequence="T" * 90,
        aa_sequence="F" * 30,
        table_id=11,
        has_start_codon=True,
        has_stop_codon=False,
    )
    
    assert orf_plus.strand == "+"
    assert orf_minus.strand == "-"
    assert orf_minus.frame == 1


def test_orf_record_all_frames() -> None:
    """Test creating ORFs in all three frames."""
    for frame in [0, 1, 2]:
        orf = OrfRecord(
            parent_id="seq1",
            orf_id=f"orf_f{frame}",
            start=frame,
            end=frame + 90,
            strand="+",
            frame=frame,
            nt_sequence="A" * 90,
            aa_sequence="K" * 30,
            table_id=11,
            has_start_codon=True,
            has_stop_codon=True,
        )
        assert orf.frame == frame


def test_orf_record_no_start_or_stop() -> None:
    """Test ORF without start or stop codons."""
    orf = OrfRecord(
        parent_id="seq1",
        orf_id="partial_orf",
        start=10,
        end=100,
        strand="+",
        frame=1,
        nt_sequence="C" * 90,
        aa_sequence="P" * 30,
        table_id=11,
        has_start_codon=False,
        has_stop_codon=False,
    )
    
    assert orf.has_start_codon is False
    assert orf.has_stop_codon is False



"""Token size estimation command."""

from typing import Optional
import traceback

try:
    import typer
except ImportError:
    typer = None


def estimate_token_size_command(
    model: str = typer.Option(
        "Rostlab/ProstT5_fp16",
        "--model",
        "-m",
        help="ProstT5 model name",
    ),
    device: Optional[str] = typer.Option(
        None,
        "--device",
        "-d",
        help="Device to use (auto/cuda/mps/cpu)",
    ),
    start_length: int = typer.Option(
        3000,
        "--start",
        "-s",
        help="Starting total protein length to test (amino acids)",
    ),
    end_length: int = typer.Option(
        10000,
        "--end",
        "-e",
        help="Maximum total protein length to test (amino acids)",
    ),
    step: int = typer.Option(
        1000,
        "--step",
        help="Increment between test lengths (amino acids)",
    ),
    num_trials: int = typer.Option(
        3,
        "--trials",
        "-t",
        help="Number of trials per length for robustness",
    ),
    base_protein_length: int = typer.Option(
        100,
        "--base-length",
        "-b",
        help="Approximate length of individual proteins (amino acids)",
    ),
) -> None:
    """Estimate optimal token size for GPU encoding.

    This command generates random protein sequences of increasing length
    and tests encoding to find the maximum that can be encoded on the
    available GPU without running out of memory.

    The recommended token size is returned as 90% of the maximum for safety.
    """
    try:
        from ...encode3di import ProstT5ThreeDiEncoder, estimate_token_size

        typer.echo("=" * 60)
        typer.echo("Token Size Estimation")
        typer.echo("=" * 60)
        typer.echo(f"Model: {model}")
        typer.echo(f"Testing range: {start_length} to {end_length} AA (step: {step})")
        typer.echo(f"Trials per length: {num_trials}")
        typer.echo(f"Base protein length: {base_protein_length} AA")
        typer.echo("=" * 60)

        typer.echo("\nInitializing ProstT5 encoder...")
        encoder = ProstT5ThreeDiEncoder(model_name=model, device=device)
        typer.echo(f"Using device: {encoder.device}")

        typer.echo("\nStarting estimation (this may take several minutes)...")
        typer.echo("Watch for Out of Memory errors to identify the limit.\n")

        results = estimate_token_size(
            encoder=encoder,
            start_length=start_length,
            end_length=end_length,
            step=step,
            num_trials=num_trials,
            base_protein_length=base_protein_length,
        )

        typer.echo("\n" + "=" * 60)
        typer.echo("RESULTS")
        typer.echo("=" * 60)
        typer.echo(f"Device: {results['device']}")
        typer.echo(f"Max successful length: {results['max_length']} amino acids")
        typer.echo(
            f"Recommended token size: {results['recommended_token_size']} amino acids"
        )
        typer.echo("\nTrials per length:")
        for length, trials in sorted(results["trials_per_length"].items()):
            status = "✓" if trials > 0 else "✗"
            typer.echo(f"  {status} {length} AA: {trials}/{num_trials} successful")
        typer.echo("=" * 60)
        typer.echo("\n✓ Estimation complete!")
        typer.echo(
            f"\nTo use this in encoding, add: --encoding-size {results['recommended_token_size']}"
        )

    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        traceback.print_exc()
        raise typer.Exit(3)
